/*-
 * Copyright (C) @BABOLO  2009 May 08 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2009 May 08 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: ipfw_get.c,v 1.8 2011/11/26 10:25:05 babolo Exp $";
#endif /* not lint */

#include <machine/param.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <mife.h>
#include <err.h>
#include <multilar.h>
#include <babolo/BLINflag.h>
#include "bsdfig.h"
#include "bsdfig_local.h"

#define BSDFIG_DOSOCKSET 1
#define BSDFIG_DOSORIGID 2

static int
dosockopt( u_int32_t flags, int s, int level, int optname
         , void *ini, size_t inl
         , char **buf, socklen_t *size, size_t delta
         ) {
    socklen_t len, try;
    u_int32_t flocal;
    int ex = EX_OK;

    if  (BSDFIG_DOSOCKSET & flags) {
        if  (!ini) {
            ifBLIN_QV1(flags) warnx("bsdfig: dosockopt: nothing to set");
            errno = EINVAL;
            ex = EX_USAGE;
        }
        if  (!ex && (ex = setsockopt(s, level, optname, ini + delta, inl - delta)) < 0) {
            ifBLIN_QV2(flags) warn( "bsdfig: dosockopt: getsockopt(%d, %d, %u, *, %"BLIN_U")"
                                  , s, level, optname, inl - delta
                                  );
            ex = EX_OSERR;
        }
    } else {
        flocal = !*buf;
        len = *size;
        if  (*size < delta) {
            ifBLIN_QV1(flags) warnx("dosockopt: size %d < delta %d", *size, delta);
            errno = EINVAL;
            ex = EX_SOFTWARE;
        }
        for (try = len; !ex; try *= 2) {
            len = try - delta;
            if  (flocal) {
                free(*buf);
                if  (!(*buf = malloc(try))) {
                    ifBLIN_QV1(flags) warnx("bsdfig: dosockopt: no mem %d", try);
                    errno = ENOMEM;
                    ex = EX_OSERR;
                    break;
            }   }
            if  (ini) {
                if  (inl > try) {
                    ifBLIN_QV2(flags) warnx("dosockopt: %"BLIN_U" in > %u out", inl, try);
                    len = len + delta;
                    continue;
                }
                bcopy(ini, *buf, inl);
            }
            if  ((ex = getsockopt(s, level, optname, *buf + delta, &len)) < 0) {
                ifBLIN_QV2(flags) warn( "dosockopt: getsockopt(%d, %d, %d, *, %"BLIN_U")"
                                      , s, level, optname, try - delta
                                      );
                ex = EX_OSERR;
                break;
            }
            len = len + delta;
            ifBLIN_QV5(flags)
                fprintf(stderr, " bsdfig: dosockopt try=%d len=%d ex=%d\n", try, len, ex);
            if  ((BSDFIG_DOSORIGID & flags) || (len < try)) break;
        }
        *size = len;
    }
    return(ex);
}

bsdfig_dummynets *
bsdfig_read_dummynets(bsdfig_conf *cnf) {
    bsdfig_dummynets *ret = NULL;
    socklen_t len;
    int sock, ex = EX_OK;
#   define FIRST_DUMMY_SIZE PAGE_SIZE

    if  (!cnf) errx(EX_USAGE, "bsdfig_read_dummynets: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_dummynets %08X\n", cnf->flags);
    if  ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        ifBLIN_QV1(cnf->flags) warn("bsdfig_read_dummynets: socket");
        ex = EX_OSERR;
    }
    len = FIRST_DUMMY_SIZE;
    if  ((ex = dosockopt( cnf->flags & BLIN_VERMASK
                        , sock, IPPROTO_IP, IP_DUMMYNET_GET
                        , NULL, 0
                        , (char **)&ret, &len, sizeof(size_t)
        ))              ) {
        ifBLIN_QV2(cnf->flags) warn("bsdfig_read_dummynets: dosockopt");
    }
    ret->size = len - sizeof(size_t);
    if  (!ex && (len < sizeof(size_t))) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_read_dummynets: len %d < %d", len, sizeof(size_t));
        errno = EINVAL;
        ex = EX_SOFTWARE;
    }
    if  (ex) {
        free(ret);
        ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_dummynets %d\n", ex);
    } else {
        ifBLIN_QV5(cnf->flags) {
            unsigned i;
            for (i = 0; i < ret->size; i++) {
                if  (!(i & 0xF)) fprintf(stderr, "\n%08X:", i);
                if  (!(i & 0x7)) fprintf(stderr, " ");
                if  (!(i & 0x1)) fprintf(stderr, " ");
                fprintf(stderr, "%02X", ((u_char *)&(ret->dummies))[i]);
            }
            fprintf(stderr, "\n");
        }
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_dummynets [%"BLIN_U"]\n", ret->size);
    }
    return(ret);
}

bsdfig_ipfwrules *
bsdfig_read_ipfwrules(bsdfig_conf *cnf) {
    bsdfig_ipfwrules *ret = NULL;
    socklen_t len;
    int sock, ex = EX_OK;
#   define FIRST_RULE_SIZE PAGE_SIZE

    if  (!cnf) errx(EX_USAGE, "bsdfig_read_ipfwrules: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_ipfwrules %08X\n", cnf->flags);
    if  ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        ifBLIN_QV1(cnf->flags) warn("bsdfig_read_ipfwrules: socket");
        ex = EX_OSERR;
    }
    len = FIRST_RULE_SIZE;
    if  ((ex = dosockopt( cnf->flags & BLIN_VERMASK
                        , sock, IPPROTO_IP, IP_FW_GET
                        , NULL, 0
                        , (char **)&ret, &len, sizeof(size_t)
        ))              ) {
        ifBLIN_QV1(cnf->flags) warn("bsdfig_read_ipfwrules: dosockopt");
    }
    ret->size = (len - sizeof(size_t)) / sizeof(ipfw_insn);
    if  (ex) {
        free(ret);
        ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwrules %d\n", ex);
    } else {
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwrules [%"BLIN_U"]\n", ret->size);
    }
    return(ret);
}

#if __FreeBSD_version >= 700000
bsdfig_ipfwnats *
bsdfig_read_ipfwnats(bsdfig_conf *cnf) {
#   define FIRST_NAT_SIZE PAGE_SIZE
    bsdfig_ipfwnats *ret = NULL;
    int sock, ex = EX_OK;
    socklen_t len;

    if  (!cnf) errx(EX_USAGE, "bsdfig_read_ipfwnats: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_ipfwnats %08X\n", cnf->flags);
    if  ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        ifBLIN_QV1(cnf->flags) warn("bsdfig_read_ipfwnats: socket");
        ex = EX_OSERR;
    }
    len = FIRST_NAT_SIZE;
    if  ((ex = dosockopt( cnf->flags & BLIN_VERMASK
                        , sock, IPPROTO_IP, IP_FW_NAT_GET_CONFIG
                        , NULL, 0
                        , (char **)&ret, &len, sizeof(size_t)
        ))              ) {
        ifBLIN_QV2(cnf->flags) warn("bsdfig_read_ipfwnats: dosockopt");
    }
    ret->size = len;
    if  (ex) {
        free(ret);
        ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwnats %d\n", ex);
    } else {
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwnats [%"BLIN_U"]\n", ret->size);
    }
    return(ret);
}

bsdfig_ipfwnatl *
bsdfig_read_ipfwnatl(bsdfig_conf *cnf) {
#   define FIRST_NAT_SIZE PAGE_SIZE
    bsdfig_ipfwnatl *ret = NULL;
    int sock, ex = EX_OK;
    socklen_t len;

    if  (!cnf) errx(EX_USAGE, "bsdfig_read_ipfwnats: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_ipfwnats %08X\n", cnf->flags);
    if  ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        ifBLIN_QV1(cnf->flags) warn("bsdfig_read_ipfwnats: socket");
        ex = EX_OSERR;
    }
    len = FIRST_NAT_SIZE;
    if  ((ex = dosockopt( cnf->flags & BLIN_VERMASK
                        , sock, IPPROTO_IP, IP_FW_NAT_GET_LOG
                        , NULL, 0
                        , (char **)&ret, &len, sizeof(size_t)
        ))              ) {
        ifBLIN_QV2(cnf->flags) warn("bsdfig_read_ipfwnats: dosockopt");
    }
    ret->size = len;
    if  (ex) {
        free(ret);
        ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwnats %d\n", ex);
    } else {
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwnats [%"BLIN_U"]\n", ret->size);
    }
    return(ret);
}
#endif

bsdfig_ipfwtables *
bsdfig_read_ipfwtables(bsdfig_conf *cnf) {
    bsdfig_ipfwtables *ret = NULL;
    int sock, ex = EX_OK;
    socklen_t len, try;
    ipfw_table tbl;
    size_t i;

    if  (!cnf) errx(EX_USAGE, "bsdfig_read_ipfwtables: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_ipfwtables %08X\n", cnf->flags);
    if  (!(ret = malloc(cnf->tabn * sizeof(ipfw_table*) + sizeof(bsdfig_ipfwtables)))) {
        ifBLIN_QV1(cnf->flags) warnx( "bsdfig_read_ipfwtables: no mem %"BLIN_U
                                    , cnf->tabn * sizeof(ipfw_table*) + sizeof(bsdfig_ipfwtables)
                                    );
        errno = ENOMEM;
        ex = EX_OSERR;
    }
    if  (!ex) {
        ret->tabn = cnf->tabn;
        bzero(&(ret->tables), ret->tabn * sizeof(ipfw_table*));
    }
    if  (!ex && (sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        ifBLIN_QV1(cnf->flags) warn("bsdfig_read_ipfwtables: socket");
        ex = EX_OSERR;
    }
    for (i = 0; !ex && (i < ret->tabn); i++) {
        try = i;
        len = sizeof(try);
        if  ((ex = getsockopt(sock, IPPROTO_IP, IP_FW_TABLE_GETSIZE, &try, &len)) < 0) {
            ifBLIN_QV1(cnf->flags) warn("bsdfig_read_ipfwtables: getsockopt");
            ex = EX_OSERR;
            break;
        }
        if  (try) {
            try = (((try + 1) * sizeof(ipfw_table_entry)) + sizeof(ipfw_table) + 15) & ~15;
            bzero(&tbl, sizeof(tbl));
            tbl.tbl = i;
            do {
                free(ret->tables[i]);
                ret->tables[i] = NULL;
                if  ((ex = dosockopt( (cnf->flags & BLIN_VERMASK) | BSDFIG_DOSORIGID
                                    , sock, IPPROTO_IP, IP_FW_TABLE_LIST
                                    , &tbl, sizeof(tbl)
                                    , (char **)&(ret->tables[i]), &try, 0
                    ))              ) {
                    ifBLIN_QV1(cnf->flags) warn("bsdfig_read_ipfwtables: dosockopt");
                    break;
                }
                try = (void*)&(ret->tables[i]->ent[ret->tables[i]->cnt + 1]) - (void*)(ret->tables[i]);
            } while ((void*)&(ret->tables[i]->ent[ret->tables[i]->cnt]) > (void*)ret->tables[i] + try);
    }   }
    if  (ex) {
        if  (ret) for (i = 0; i < ret->tabn; i++) free(ret->tables[i]);
        free(ret);
        ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwtables %d\n", ex);
    } else {
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwtables [%"BLIN_U"]\n", ret->tabn);
    }
    return(ret);
}

static const size_t mule[3] = {1024U, 512U, 512U};

bsdfig_conf *
bsdfig_init(u_int32_t flags, FILE *out, bsdfig_conf *slug) {
    bsdfig_conf *cnf;
    int ex = EX_OK, nv;
    size_t ns, i;

    ifBLIN_QV4(flags) fprintf(stderr, "+bsdfig_init %08X\n", flags);
    do {
        if  (slug) {
            cnf = slug;
        } else if (!(cnf = calloc(1, sizeof(bsdfig_conf)))) {
            ifBLIN_QV1(flags) warnx("bsdfig_init: no init mem %"BLIN_U, sizeof(bsdfig_conf));
            errno = ENOMEM;
            ex = EX_OSERR;
            break;
        } else cnf->flags = flags;

        if  (!cnf->out) cnf->out = out;
        if  (!cnf->syntax) cnf->syntax = &bsdfig_s;
        if  (!cnf->fibn) {
            ns = sizeof(int);
            if  (sysctlbyname("net.fibs", &nv, &ns, NULL, 0) < 0) {
                ifBLIN_QV1(cnf->flags) warn("bsdfig_init: sysctlbyname");
                ex = EX_OSERR;
                break;
            }
            cnf->fibn = nv;
        }
        if  (!cnf->tabn) cnf->tabn = IPFW_TABLES_MAX;

        if  (!(BSDFIG_0CONF & cnf->flags)) break;

        if  (!cnf->conf) {
            if  (!( cnf->conf
                  = calloc(1, (cnf->tabn - 1) * sizeof(mular_descriptor*) + sizeof(bsdfig_allconf))
                ) ) {
                ifBLIN_QV1(flags) warnx("bsdfig_init: no mem %"BLIN_U, sizeof(bsdfig_allconf));
                errno = ENOMEM;
                ex = EX_OSERR;
                break;
            } else cnf->conf->flags = cnf->flags;
        }

        if  (!cnf->conf->fibn) cnf->conf->fibn = cnf->fibn;
        if  (!cnf->conf->tabn) cnf->conf->tabn = cnf->tabn;

        if  (!cnf->conf->dummynets) {
            if  (!(cnf->conf->dummynets = mular_create(0, 3, 0, mule))) {
                ifBLIN_QV1(flags) warn("bsdfig_init: mular_create for dummynets");
                ex = EX_SOFTWARE;
                break;
            } else cnf->conf->dummynets->tofree  = free;
        }
      
        if  (!cnf->conf->ipfwrules) {
            if  (!(cnf->conf->ipfwrules = mular_create(0, 3, 0, mule))) {
                ifBLIN_QV1(flags) warn("bsdfig_init: mular_create for ipfwrules");
                ex = EX_SOFTWARE;
                break;
            } else cnf->conf->ipfwrules->tofree  = free;
        }

        if  (!cnf->conf->ipfwtables) {
            if  (!( cnf->conf->ipfwtables
                  = calloc(1, cnf->conf->tabn * sizeof(mular_descriptor*))
                ) ) {
                ifBLIN_QV1(flags)
                    warnx("bsdfig_init: no mem %"BLIN_U, cnf->conf->tabn * sizeof(mular_descriptor*));
                errno = ENOMEM;
                ex = EX_OSERR;
                break;
        }   }

        for (i = 0; !ex && (i < cnf->conf->tabn); i++) {
            if  (!cnf->conf->ipfwtables[i]) {
                if  (!(cnf->conf->ipfwtables[i] = mular_create(0, 3, 0, mule))) {
                    ifBLIN_QV1(flags) warn("bsdfig_init: mular_create for ipfwtables");
                    ex = EX_SOFTWARE;
                    break;
                } else cnf->conf->ipfwtables[i]->tofree  = free;
        }   }

        if  (0 && !cnf->conf->routes) { /* No routes yet */
            if  (!( cnf->conf->routes
                  = calloc(1, cnf->conf->fibn * sizeof(void*))
                ) ) {
                ifBLIN_QV1(flags)
                    warnx("bsdfig_init: no mem %"BLIN_U, cnf->conf->fibn * sizeof(void*));
                errno = ENOMEM;
                ex = EX_OSERR;
                break;
        }   }

        if  (ex || !(BSDFIG_MYCNF & flags)) break;
        for (i = 0; i < (cnf->tabn / BSDFIG_BASELEN); i++) cnf->conf->my_ipfwtables[i] = ~0LLU;
        for (i = 0; i < (cnf->tabn % BSDFIG_BASELEN); i++) {
            cnf->conf->my_ipfwtables[cnf->tabn / BSDFIG_BASELEN] |= 1LLU << i;
        }
        memset(&(cnf->conf->my_pipes    ), ~0, BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BYTELEN);
        memset(&(cnf->conf->my_queues   ), ~0, BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BYTELEN);
        memset(&(cnf->conf->my_ipfwrules), ~0, BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BYTELEN);
        cnf->conf->my_pipes    [0] = 0xFFFFFFFFFFFFFFFELLU;
        cnf->conf->my_queues   [0] = 0xFFFFFFFFFFFFFFFELLU;
        cnf->conf->my_ipfwrules[0] = 0xFFFFFFFFFFFFFFFELLU;
        cnf->conf->my_ipfwrules[(BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BASELEN) - 1] = 0x7FFFFFFFFFFFFFFFLLU;
    } while (0);
    ifBLIN_QV4(flags) fprintf(stderr, "-bsdfig_init %d\n", ex);
    return(cnf);
}
