/*-
 * Copyright (C) @BABOLO  2009 May 08 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2009 May 08 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: ipfw_out.c,v 1.30 2011/11/26 12:29:09 babolo Exp $";
#endif /* not lint */

#include <machine/param.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <mife.h>
#include <err.h>
#include <multilar.h>
#include <babolo/BLINflag.h>
#include "bsdfig.h"
#include "bsdfig_local.h"

#define SYNCO(x)   (cnf->syntax->p[x])
#define DOIT(ex)   ((ex) <= (BSDFIG_EXLEM & cnf->flags))
#define STYLEIS(x) (BSDFIG_ARGTY(x) & SYNCO(opcode).P)
#define STYLAR2(x, y) ((BSDFIG_ARGTY(x) | BSDFIG_ARGTY(y)) & SYNCO(opcode).P)
#define STYLAR3(x, y, z) ((BSDFIG_ARGTY(x) | BSDFIG_ARGTY(y) | BSDFIG_ARGTY(z)) & SYNCO(opcode).P)

                        /* flags for flocal          */
#define DSABLQ 0x100000 /* print disable             */
#define DSABLE 0x080000 /* print disable             */
#define FPRINT 0x040000 /* overall debug             */
#define RPRINT 0x020000 /* in rule debug             */
#define OPRINT 0x010000 /* in op debug               */
#define QPRINT 0x070000 /* debug print enable        */
#define PRECHE 0x008000 /* not first pre char enable */
#define LFIRST 0x004000 /* first pre char in list    */
#define SRUEND 0x002000 /* last static rule found    */
#define ACTION 0x001000 /* action rule found         */

#define MSG(E, S, N1, N2) do {                                                                        \
    ifBLIN_QV##E(cnf->flags) warnx( "" S, SYNCO(opcode).opn, cmd->opcode, N1, N2);                    \
} while(0)

#define TOBREAK(E) {                                                                                  \
    ex = (E);                                                                                         \
    errno = EINVAL;                                                                                   \
    break;                                                                                            \
}

#define BREAKORNOT(E) {                                                                               \
    ex = (E);                                                                                         \
    errno = 0;                                                                                        \
    if  (!DOIT(ex)) break;                                                                            \
}

#define TOBREAKORNOT(E) {                                                                             \
    BREAKORNOT(E)                                                                                     \
    flocal |= OPRINT;                                                                                 \
}

#define I4H(X) (((X) >> 24) & 0xFF), (((X) >> 16) & 0xFF), (((X) >> 8) & 0xFF), ((X) & 0xFF)
#define I4N(X) ((X) & 0xFF), (((X) >> 8) & 0xFF), (((X) >> 16) & 0xFF), (((X) >> 24) & 0xFF)
#define I62(X) (ntohs((X)->__u6_addr.__u6_addr16[0] & 0xFFFF))                                        \
             , (ntohs((X)->__u6_addr.__u6_addr16[1] & 0xFFFF))                                        \
             , (ntohs((X)->__u6_addr.__u6_addr16[2] & 0xFFFF))                                        \
             , (ntohs((X)->__u6_addr.__u6_addr16[3] & 0xFFFF))                                        \
             , (ntohs((X)->__u6_addr.__u6_addr16[4] & 0xFFFF))                                        \
             , (ntohs((X)->__u6_addr.__u6_addr16[5] & 0xFFFF))                                        \
             , (ntohs((X)->__u6_addr.__u6_addr16[6] & 0xFFFF))                                        \
             , (ntohs((X)->__u6_addr.__u6_addr16[7] & 0xFFFF))
#define I64(X) (ntohs((X)->__u6_addr.__u6_addr32[0])), (ntohs((X)->__u6_addr.__u6_addr32[1]))         \
             , (ntohs((X)->__u6_addr.__u6_addr32[2])), (ntohs((X)->__u6_addr.__u6_addr32[3]))

#define ELPRINTO(C, E, F)                                                                             \
    elprinto elprinto;                                                                                \
    elprinto.cnf = (C);                                                                               \
    elprinto.ex = (E);                                                                    \
    elprinto.flocal = (F);

#ifdef ELPRINTO /* El Printo primitivo */

typedef struct elprinto {
    bsdfig_conf *cnf;
    unsigned    *ex;
    u_int32_t   *flocal;
} elprinto;

static int
elprint( elprinto elprinto, u_int32_t flocal, char *error, char *format, ...) {
    va_list ap;

    va_start(ap, format);
    if  (  (*(elprinto.ex) <= (BSDFIG_EXLEM & elprinto.cnf->flags))
        && !(DSABLE & *(elprinto.flocal))
        && (QPRINT & (*(elprinto.flocal) | flocal))
        && elprinto.cnf->out
        && format
        && (vfprintf(elprinto.cnf->out, format, ap) < 0)
        ) {
        ifBLIN_QV1(elprinto.cnf->flags)
            warn("bsdfig: %s: elprint: vfprintf in format %s", error, format);
        *(elprinto.ex) = EX_IOERR;
    }
    return((*(elprinto.ex) > (BSDFIG_EXLEM & elprinto.cnf->flags)) ? *(elprinto.ex) : 0);
}

static int
elabor(elprinto elprinto, u_int32_t flocal, char *error, char *format, ...) {
    va_list ap;
    int c;

    va_start(ap, format);
    if      (LFIRST & *(elprinto.flocal)) c = '(';
    else if (PRECHE & *(elprinto.flocal)) c = ' ';
    else c = 0;
    if  (c) elprint(elprinto, flocal, error, "%c", c);
    if  (  (*(elprinto.ex) <= (BSDFIG_EXLEM & elprinto.cnf->flags))
        && !(DSABLE & *(elprinto.flocal))
        && (QPRINT & (*(elprinto.flocal) | flocal))
        && elprinto.cnf->out
        && format
        && (vfprintf(elprinto.cnf->out, format, ap) < 0)
        ) {
        ifBLIN_QV1(elprinto.cnf->flags)
            warn("bsdfig: %s: elabor: vfprintf in format %s", error, format);
        *(elprinto.ex) = EX_IOERR;
    }
    *(elprinto.flocal) &= ~LFIRST;
    *(elprinto.flocal) |= PRECHE;
    return((*(elprinto.ex) > (BSDFIG_EXLEM & elprinto.cnf->flags)) ? *(elprinto.ex) : 0);
}

#define ELSTART do { *(elprinto.flocal) |=   PRECHE | LFIRST;  } while (0)
#define ELCONT  do { *(elprinto.flocal) &= ~(PRECHE | LFIRST); } while (0)

#define ELSTOP(QPRINT, ERROR)                                                                         \
    if (!(LFIRST & *(elprinto.flocal)) && elprint(elprinto, QPRINT, ERROR, ")")) break;

#endif /* ELPRINTO   El Printo primitivo */

static int
elist(bsdfig_conf *cnf, u_int64_t *larr, size_t arrl) {
    int ex = EX_OK;

    return(ex);
}

static int
elflowid(elprinto elprinto, char *error, struct ipfw_flow_id *f) {
    elprint( elprinto, QPRINT, error
#if __FreeBSD_version >= 500000
           , "%u.%u.%u.%u %u.%u.%u.%u %u %u %u %u %02X %u"
             " %X:%X:%X:%X:%X:%X:%X:%X %X:%X:%X:%X:%X:%X:%X:%X %u %u"
#else
           , "%u.%u.%u.%u %u.%u.%u.%u %u %u %u %02X"
#endif
           , I4H(f->dst_ip)    , I4H(f->src_ip)
           , f->dst_port       , f->src_port
#if __FreeBSD_version >= 500000
           , f->fib
#endif
           , f->proto
#if __FreeBSD_version >= 800504
           , f->_flags
#else
           , f->flags
#endif
#if __FreeBSD_version >= 500000
           , f->addr_type
           , I62(&(f->dst_ip6)), I62(&(f->src_ip6))
           , f->flow_id6
# if __FreeBSD_version >= 800504
           , f->extra
# else
           , f->frag_id6
# endif
#endif
           );
    return((*(elprinto.ex) > (BSDFIG_EXLEM & elprinto.cnf->flags)) ? *(elprinto.ex) : 0);
}

static int
elflowidd(elprinto elprinto, char *error, struct ipfw_flow_id *f) {
    elprint( elprinto, 0, error
#if __FreeBSD_version >= 500000
           , "%08X %08X %u %u %u %u %02X %u %08X%08X%08X%08X %08X%08X%08X%08X %u %u"
#else
           , "%08X %08X %u %u %u %02X %u"
#endif
           , f->dst_ip         , f->src_ip         , f->dst_port, f->src_port
#if __FreeBSD_version >= 500000
           , f->fib
#endif
           , f->proto
#if __FreeBSD_version >= 800504
           , f->_flags
#else
           , f->flags
#endif
#if __FreeBSD_version >= 500000
           , f->addr_type
           , I64(&(f->dst_ip6)), I64(&(f->src_ip6))
           , f->flow_id6
# if __FreeBSD_version >= 800504
           , f->extra
# else
           , f->frag_id6
# endif
#endif
           );
    return((*(elprinto.ex) > (BSDFIG_EXLEM & elprinto.cnf->flags)) ? *(elprinto.ex) : 0);
}

static int
elflowqueue(elprinto elprinto, struct dn_flow_queue **fqp, void *duend, size_t sz) {
    size_t i;

    ifBLIN_QV5(elprinto.cnf->flags) fprintf(stderr, "+elflowqueue %d\n", (int)sz);
    for (i = 0; !*(elprinto.ex) && (i < sz); (*fqp)++, i++) {
        *(elprinto.flocal) &= ~OPRINT;
        if  ((void*)((*fqp) + 1) > duend) {
            ifBLIN_QV1(elprinto.cnf->flags) warnx("bsdfig: elflowqueu: lost dn_flow_queue size");
            errno = EINVAL;
            *(elprinto.ex) = EX_DATAERR;
            break;
        }
        if  ((DSABLE & *(elprinto.flocal))) continue;
        if  (elprint(elprinto, QPRINT, "bsdfig: elflowqueu #1", "((")) break;
        if  (elflowid(elprinto, "bsdfig: elflowqueu #2", &((*fqp)->id))) break;
        if  (elprint( elprinto, QPRINT, "bsdfig: elflowqueu #3"
                    , ") %u %u %lu %llu %llu %u %u %llu %llu %llu)\n"
                    , (*fqp)->len, (*fqp)->len_bytes, (*fqp)->numbytes
                    , (*fqp)->tot_pkts, (*fqp)->tot_bytes, (*fqp)->drops
#if __FreeBSD_version > 800099
                    , (*fqp)->idle_time
#else
                    , (*fqp)->q_time
#endif
                    , (*fqp)->sched_time, (*fqp)->S, (*fqp)->F
            )       ) break;
        if  (!(QPRINT & *(elprinto.flocal))) continue;
        if  (elprint(elprinto, 0, "bsdfig: elflowqueu #4", "[(")) break;
        if  (elflowidd(elprinto, "bsdfig: elflowqueu #5", &((*fqp)->id))) break;
        if  (elprint( elprinto, 0, "bsdfig: elflowqueu #6"
                    , ") .. %u %u %lu %llu %llu %u %d %d %d %d %u . %d (%llu) (%llu) (%llu)]\n"
                    , (*fqp)->len, (*fqp)->len_bytes, (*fqp)->numbytes
                    , (*fqp)->tot_pkts, (*fqp)->tot_bytes, (*fqp)->drops, (*fqp)->hash_slot
                    , (*fqp)->avg, (*fqp)->count, (*fqp)->random
#if __FreeBSD_version > 800099
                    , (*fqp)->idle_time
#else
                    , (*fqp)->q_time
#endif
                    , (*fqp)->heap_pos
                    , (*fqp)->sched_time, (*fqp)->S, (*fqp)->F
            )       ) break;
    }
    ifBLIN_QV5(elprinto.cnf->flags) fprintf(stderr, "-elflowqueue %u\n", *(elprinto.ex));
    return((*(elprinto.ex) > (BSDFIG_EXLEM & elprinto.cnf->flags)) ? *(elprinto.ex) : 0);
}

static int
elqtail(elprinto elprinto, struct dn_flow_set *queue) {
    do {
        if  (elprint( elprinto, QPRINT, "bsdfig: elqtail: #1"
                    , " %d %f ("
                    , queue->qsize
                    , (1.0 * queue->plr) / 0x7FFFFFFF
            )       ) break;
        if  (elflowid(elprinto, "bsdfig: elqtail: #2", &(queue->flow_mask))) break;
        if  (elprint( elprinto, QPRINT, "bsdfig: elqtail: #3"
                    , ") %d %d %d %d %d %d)\n"
                    , queue->w_q
                    , queue->max_th
                    , queue->min_th
                    , queue->max_p
                    , queue->lookup_step
                    , queue->lookup_weight
            )       ) break;
    } while (0);
    return((*(elprinto.ex) > (BSDFIG_EXLEM & elprinto.cnf->flags)) ? *(elprinto.ex) : 0);
}

static int
elqtaild(elprinto elprinto, struct dn_flow_set *queue) {
    do {
        if  (elprint( elprinto, 0, "bsdfig: elqtaild: #1"
                    , " %05u %04X . %05u %d %d %d ("
                    , queue->fs_nr
                    , queue->flags_fs
                    , queue->parent_nr
                    , queue->weight
                    , queue->qsize
                    , queue->plr
            )       ) break;
        if  (elflowidd(elprinto, "bsdfig: elqtaild: #2", &(queue->flow_mask))) break;
        if  (elprint( elprinto, 0, "bsdfig: elqtaild: #3"
                    , ") %d %d . %u %d %d %d %d %d %u %u %u %u %"BLIN_X" %u %d %d %d %d)]\n"
                    , queue->rq_size
                    , queue->rq_elements
                    , queue->last_expired
                    , queue->backlogged
                    , queue->w_q, queue->max_th, queue->min_th, queue->max_p
                    , queue->c_1, queue->c_2   , queue->c_3   , queue->c_4
                    , BLIN_I(queue->w_q_lookup)
                    , queue->lookup_depth
                    , queue->lookup_step
                    , queue->lookup_weight
                    , queue->avg_pkt_size
                    , queue->max_pkt_size
            )       ) break;
    } while (0);
    return((*(elprinto.ex) > (BSDFIG_EXLEM & elprinto.cnf->flags)) ? *(elprinto.ex) : 0);
}

int
bsdfig_out_pipe(bsdfig_conf *cnf, struct dn_pipe *pipe) {
    unsigned ex = EX_OK;
    u_int32_t flocal = 0;
    ELPRINTO(cnf, &ex, &flocal);

    if  (!cnf) errx(EX_USAGE, "bsdfig_out_pipe: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_out_pipe %08X\n", cnf->flags);
    ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
    if  ((BSDFIG_PPIPE & cnf->flags) && cnf->out) flocal &= ~DSABLE; else flocal |= DSABLE;
    do {
        if  (elprint( elprinto, QPRINT, "bsdfig_out_pipe #1"
                    , (BLIN_VER1 & cnf->flags)
                    ? "PIPE(%05u %u[bit/s] %u[ms] %s (%04X"
                    : "PIPE(%05u %u %u %s (%04X"
                    , pipe->pipe_nr
                    , pipe->bandwidth
                    , pipe->delay
/* XXXX */          , pipe->if_name
                    , pipe->fs.flags_fs
            )       ) break;
        if  (elqtail(elprinto, &(pipe->fs))) break;
        if  (elprint( elprinto, 0, "bsdfig_out_pipe #2"
                    , "[%"BLIN_X" %05u %u %u .. "
                      "(%d %d %d %"BLIN_X")(%d %d %d %"BLIN_X")(%d %d %d %"BLIN_X")"
                      " %llu %d %d %llu %s %"BLIN_X" %d (."
#if __FreeBSD_version >= 500000
                    , BLIN_I(SLIST_NEXT(pipe, next))
#else
                    , pipe->next
#endif
                    , pipe->pipe_nr, pipe->bandwidth, pipe->delay
                    , pipe->scheduler_heap.size     , pipe->scheduler_heap.elements
                    , pipe->scheduler_heap.offset   , BLIN_I(pipe->scheduler_heap.p)
                    , pipe->not_eligible_heap.size  , pipe->not_eligible_heap.elements
                    , pipe->not_eligible_heap.offset, BLIN_I(pipe->not_eligible_heap.p)
                    , pipe->idle_heap.size          , pipe->idle_heap.elements
                    , pipe->idle_heap.offset        , BLIN_I(pipe->idle_heap.p)
                    , pipe->V
                    , pipe->sum
                    , pipe->numbytes
                    , pipe->sched_time
/* XXXX */          , pipe->if_name
                    , BLIN_I(pipe->ifp)
                    , pipe->ready
            )       ) break;
        if  (elqtaild(elprinto, &(pipe->fs))) break;
    } while (0);
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_out_pipe %u\n", ex);
    return(ex);
}

int
bsdfig_out_queue(bsdfig_conf *cnf, struct dn_flow_set *queue) {
    unsigned ex = EX_OK;
    u_int32_t flocal = 0;
    ELPRINTO(cnf, &ex, &flocal);

    if  (!cnf) errx(EX_USAGE, "bsdfig_out_queue: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_out_queue %08X\n", cnf->flags);
    ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
    if  ((BSDFIG_QUEUE & cnf->flags) && cnf->out) flocal &= ~DSABLE; else flocal |= DSABLE;
    do {
        if  (elprint( elprinto, QPRINT, "bsdfig_out_queue #1"
                    , "QUEUE(%05u %04X %05d %d"
                    , queue->fs_nr
                    , queue->flags_fs
            )       ) break;
        if  (elqtail(elprinto, queue)) break;
        if  (elprint( elprinto, 0, "bsdfig_out_queue #2"
                    , "[%"BLIN_X
#if __FreeBSD_version >= 500000
                    , BLIN_I(SLIST_NEXT(queue, next))
#else
                    , queue->next
#endif
            )       ) break;
        if  (elqtaild(elprinto, queue)) break;
    } while (0);
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_out_queue %u\n", ex);
    return(ex);
}

int
ckout_dummynets(bsdfig_conf *cnf, u_int32_t kind, bsdfig_dummynets *dummies) {
    struct dn_flow_queue *fq;
    u_int32_t flocal = 0;
    void *dummy, *duend;
    unsigned ex = EX_OK;
    ELPRINTO(cnf, &ex, &flocal);

    ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
    for (dummy = &(dummies->dummies), duend = dummy + dummies->size; dummy < duend; dummy = fq) {
        flocal &= ~(OPRINT | RPRINT);
        if  ((void*)((u_int32_t**)dummy + 1) > duend) {
            ifBLIN_QV1(cnf->flags) warnx("bsdfig: ckout_dummynets: lost header size");
            errno = EINVAL;
            ex = EX_DATAERR;
            break;
        }
        if  (BLIN_I(*(u_int32_t**)dummy) == DN_IS_PIPE) {
            struct dn_pipe *pipe;

            flocal |= DSABLE;
            pipe = dummy;
            if  ((void*)(pipe + 1) > duend) {
                ifBLIN_QV1(cnf->flags) warnx("bsdfig: ckout_dummynets: lost pipe size");
                errno = EINVAL;
                ex = EX_DATAERR;
                break;
            }
            if  (DN_IS_PIPE == kind) {
                if  ((ex = bsdfig_out_pipe(cnf, pipe))) break;
                if  ((BSDFIG_DPIPE & cnf->flags) && cnf->out) flocal &= ~DSABLE;
            }
            fq = (struct dn_flow_queue *)(pipe + 1);
            if  (elflowqueue(elprinto, &fq, duend, pipe->fs.rq_elements)) break;
        } else if (BLIN_I(*(u_int32_t**)dummy) == DN_IS_QUEUE) {
            struct dn_flow_set *queue;

            flocal |= DSABLE;
            queue = dummy;
            if  ((void*)(queue + 1) > duend) {
                ifBLIN_QV1(cnf->flags) warnx("bsdfig: ckout_dummynets: lost queue size");
                errno = EINVAL;
                ex = EX_DATAERR;
                break;
            }
            if  (DN_IS_QUEUE == kind) {
                if  ((ex = bsdfig_out_queue(cnf, queue))) break;
                if  ((BSDFIG_DQUEU & cnf->flags) && cnf->out) flocal &= ~DSABLE;
            }
            fq = (struct dn_flow_queue *)(queue + 1);
            if  (elflowqueue(elprinto, &fq, duend, queue->rq_elements)) break;
        } else {
            ifBLIN_QV1(cnf->flags) warnx( "bsdfig: ckout_dummynets: unknown data type @%"BLIN_X"=%08X"
                                        , BLIN_I(dummy - (void*)&(dummies->dummies))
                                        , BLIN_I(*(u_int32_t**)dummy)
                                        );
            errno = EINVAL;
            ex = EX_DATAERR;
            break;
    }   }
    return(ex);
}

int
bsdfig_ckout_pipes(bsdfig_conf *cnf, bsdfig_dummynets *dummies) {
    unsigned ex = EX_OK;

    if  (!cnf) errx(EX_USAGE, "bsdfig_ckout_pipes: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ckout_pipes %08X\n", cnf->flags);
    if  (!dummies) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_ckout_pipes: no rules");
        ex = EX_DATAERR;
        errno = EINVAL;
    } else {
        ex = ckout_dummynets(cnf, DN_IS_PIPE, dummies);
    }
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_ckout_pipes %u\n", ex);
    return(ex);
}

int
bsdfig_ckout_queues(bsdfig_conf *cnf, bsdfig_dummynets *dummies) {
    unsigned ex = EX_OK;

    if  (!cnf) errx(EX_USAGE, "bsdfig_ckout_queues: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ckout_queues %08X\n", cnf->flags);
    if  (!dummies) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_ckout_queues: no rules");
        ex = EX_DATAERR;
        errno = EINVAL;
    } else {
        ex = ckout_dummynets(cnf, DN_IS_QUEUE, dummies);
    }
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_ckout_queues %u\n", ex);
    return(ex);
}

int
bsdfig_ckout_ipfwrule(bsdfig_conf *cnf, struct ip_fw *rule) {
    size_t rupnt = 0;
    u_int32_t flocal = 0;
    unsigned ex = EX_OK;
    ELPRINTO(cnf, &ex, &flocal);

    if  (!cnf) errx(EX_USAGE, "bsdfig_ckout_ipfwrule: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ckout_ipfwrule %08X\n", cnf->flags);
    if  (!((BSDFIG_PRUFW & cnf->flags) && (cnf->out))) flocal |= DSABLE;
    ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
    if  (!rule) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_ckout_ipfwrule: no rule");
        ex = EX_DATAERR;
        errno = EINVAL;
    }
    do {
        size_t cmlen, cmpnt, cclas;

        if  (!DOIT(ex) || !(rupnt < BSDFIG_RULENHEAD + rule->cmd_len)) break;
        flocal &= ~(OPRINT | RPRINT | ACTION);
        if  (elprint( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #1", "RULE(%05u %02u"
                    , rule->rulenum, rule->set
            )       ) break;
        for (cmpnt = 0, cclas = 0; DOIT(ex) && (cmpnt < rule->cmd_len); cmpnt += cmlen) {
            bsdfig_syllable *s;
            u_int32_t opcode;
            ipfw_insn *cmd;
            size_t inlen;
        /*********************************************************************************************
        ;*                                                                                           *
        ;*                                                        *
        ;*                                                                                           *
        ;*********************************************************************************************/
        ;   flocal &= ~OPRINT;
        ;   cmd = &(rule->cmd[cmpnt]);
        ;   s = (bsdfig_syllable*)cmd;
        ;   cmlen = F_LEN(cmd);
        ;   if  (cmpnt + cmlen > rupnt + BSDFIG_RULENHEAD + rule->cmd_len) {
        ;       ifBLIN_QV1(cnf->flags)
                    warnx( "bsdfig_ckout_ipfwrule: no space for command of lenght %"
                           BLIN_U" @ %"BLIN_U":%"BLIN_U
                         , BLIN_I(cmlen)
                         , BLIN_I(rupnt)
                         , BLIN_I(cmpnt)
                         );
        ;       TOBREAK(EX_DATAERR);
        ;   }
        ;   if  ((opcode = cmd->opcode) >= O_LAST_OPCODE) {
        ;       opcode = O_LAST_OPCODE;
        ;       ifBLIN_QV2(cnf->flags)
                    warnx( "bsdfig_ckout_ipfwrule: illegal opcode %d >= LAST_OPCODE %d\n"
                         , cmd->opcode
                         , O_LAST_OPCODE
                         );
        ;       TOBREAKORNOT(BSDFIG_EXLES);
        ;   }
        ;   if  (cmlen < SYNCO(opcode).l) {
        ;       MSG(1, "length of %s[%d] is %"BLIN_U" < %d", BLIN_I(cmlen), SYNCO(opcode).l);
        ;       TOBREAK(EX_DATAERR);
        ;   }
        ;   if  (cmlen > SYNCO(opcode).L) {
        ;       MSG(2, "length of %s[%d] is %"BLIN_U" > %d", BLIN_I(cmlen), SYNCO(opcode).L);
        ;       TOBREAKORNOT(BSDFIG_EXLEE);
        ;   }
        ;   if  ((BSDFIG_LMDLM & SYNCO(opcode).M) && !(cmlen & 1)) {
        ;       MSG(2, "length of %s[%d] is %"BLIN_U" not %dn+1", cmlen, 2);
        ;       TOBREAKORNOT(BSDFIG_EXLEI);
        ;   }
        ;   if  ((BSDFIG_LMDLI & SYNCO(opcode).M) && ((cmlen - 1) & 7)) {
        ;       MSG(2, "length of %s[%d] is %"BLIN_U" not %dn+1", cmlen, 8);
        ;       TOBREAKORNOT(BSDFIG_EXLEW);
        ;   }
        ;   if  (STYLEIS(BSDFIG_A1NUM)) {
        ;       inlen = cmd->arg1;
        ;       if  (cmlen < (inlen + 1)) {
        ;           MSG(1, "length of %s[%d] is %"BLIN_U" < arg1(%d)+1", cmlen, cmd->arg1);
        ;           TOBREAK(EX_DATAERR);
        ;       } else if (cmlen > (inlen + 1)) {
        ;           MSG(2, "length of %s[%d] is %"BLIN_U" > arg1(%d)+1", cmlen, cmd->arg1);
        ;           TOBREAKORNOT(BSDFIG_EXLEE);
        ;       }
        ;   } else if (STYLEIS(BSDFIG_A1BIT)) {
        ;       inlen = (cmd->arg1 + 31) / 32;
        ;       if  (cmlen < (inlen + 2)) {
        ;           MSG(1, "length of %s[%d] is %"BLIN_U" < arg1(%d)/32+2", cmlen, cmd->arg1);
        ;           TOBREAK(EX_DATAERR);
        ;       } else if (cmlen > (inlen + 2)) {
        ;           MSG(2, "length of %s[%d] is %"BLIN_U" > arg1(%d)/32+2", cmlen, cmd->arg1);
        ;           TOBREAKORNOT(BSDFIG_EXLEE);
        ;       }
        ;   } else {
        ;       inlen = 64;
        ;   }
        ;   if  (BSDFIG_ROLAn <= cclas) {
        ;       MSG(2, "no more cclas for %s[%d]: %d <= %"BLIN_U, BSDFIG_ROLAa, cclas);
        ;       BREAKORNOT(BSDFIG_EXLEE);
        ;   }
        ;   if  ((BSDFIG_ROLAA & SYNCO(opcode).R) < (BSDFIG_ROLAA & cclas)) {
        ;       /*      ,   act_ofs,    */
        ;       MSG( 2, "cclas of %s[%d] is %02X < %02X - out of order after act_ofs"
                   , SYNCO(opcode).R
                   , cclas
                   );
        ;       BREAKORNOT(BSDFIG_EXLEI);
        ;   } else if ((BSDFIG_ROLAm & SYNCO(opcode).R) < (cclas & BSDFIG_ROLAm)) {
        ;       /*      */
        ;       MSG(2, "cclas of %s[%d] is %02X < %02X - out of order #1", SYNCO(opcode).R, cclas);
        ;       BREAKORNOT(BSDFIG_EXLEW);
        ;   } else if ((BSDFIG_ROLAm & SYNCO(opcode).R) < cclas) {
        ;       /*      */
        ;       MSG(2, "cclas of %s[%d] is %02X <= %02X - out of order #2", SYNCO(opcode).R, cclas);
        ;       BREAKORNOT(BSDFIG_EXLEW);
        ;   }
        ;   cclas = SYNCO(opcode).R;
        ;   if  (cmpnt == rule->act_ofs) {
        ;       if  (elprint(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #2", " +")) break;
        ;       flocal |= ACTION;
        ;   } else if (!(ACTION & flocal) && (cmpnt > rule->act_ofs)) {
        ;       if  (elprint(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #3", " *")) break;
        ;       flocal |= ACTION;
        ;       ifBLIN_QV2(cnf->flags) warnx("bsdfig: bogus act_ofs in rule @ %"BLIN_U, rupnt);
        ;       BREAKORNOT(BSDFIG_EXLEE);
        ;   }
        ;   if  ((cclas & BSDFIG_ROLAA) && !(ACTION & flocal)) {
        ;       /* XXXX    ,  act_ofs  , */
        ;       /*                       */
        ;       MSG(2, "cclas of %s[%d] is %"BLIN_U" >= %d but no act_ofs", cclas, BSDFIG_ROLAa);
        ;       BREAKORNOT(BSDFIG_EXLEW);
        ;   }
        /*********************************************************************************************
        ;*                                                                                           *
        ;*                                     *
        ;*                                                                                           *
        ;*********************************************************************************************/
        ;   if  (elprint(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #4"
                , (F_NOT & cmd->len) ? " ~%s" : " %s", SYNCO(opcode).opn)
                ) break;
        ;   if  (elprint(elprinto, 0, "bsdfig_ckout_ipfwrule #5", "[%02X]", cmd->opcode)) break;
        ;   ELSTART;
        ;   if  (STYLAR3(BSDFIG_A1NUM, BSDFIG_A1BIT, BSDFIG_A163K)) {
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #6", "%u", cmd->arg1)) break;
        ;   }
        ;   if  (STYLEIS(BSDFIG_A1255)) {
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #7", "%u", cmd->arg1)) break;
        ;       if  (cmd->arg1 > 255) {
        ;           MSG(2, "arg1 in %s[%d] is %d > %d", cmd->arg1, 255);
        ;           BREAKORNOT(BSDFIG_EXLEE);
        ;   }   }
        ;   if  (STYLEIS(BSDFIG_A1256)) {
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #8", "%u", cmd->arg1)) break;
        ;       if  ((cmd->arg1 > 256) || (cmd->arg1 < 2)) {
        ;           MSG(2, "arg1 in %s[%d] is %d not in 2..%d", cmd->arg1, 256);
        ;           BREAKORNOT(BSDFIG_EXLEE);
        ;   }   }
        ;   if  (STYLEIS(BSDFIG_A1127)) {
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #9", "%u", cmd->arg1)) break;
        ;       if  (cmd->arg1 > IPFW_TABLES_MAX) {
        ;           MSG(2, "arg1 in %s[%d] is %d > FIB number %d", cmd->arg1, IPFW_TABLES_MAX);
        ;           BREAKORNOT(BSDFIG_EXLEE);
        ;   }   }
        ;   if  (STYLEIS(BSDFIG_A1FIB)) {
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #10", "%u", cmd->arg1)) break;
        ;       if  (cmd->arg1 > cnf->fibn) {
        ;           MSG(2, "arg1 in %s[%d] is %d > FIB number %d", cmd->arg1, cnf->fibn);
        ;           BREAKORNOT(BSDFIG_EXLEE);
        ;   }   }
        ;   if  (STYLEIS(BSDFIG_A1_3)) {
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #11", "%u", cmd->arg1)) break;
        ;       if  (cmd->arg1 > 3) {
        ;           MSG(2, "arg1 in %s[%d] is %d > %d", cmd->arg1, 3);
        ;           BREAKORNOT(BSDFIG_EXLEE);
        ;   }   }
        ;   if  (  (STYLEIS(BSDFIG_A1_BB))
                && elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #12"
                         , "0x%02X 0x%02X", cmd->arg1 & 0xFF, (cmd->arg1 >> 8) & 0xFF
                )        ) break;
        ;   if  (STYLEIS(BSDFIG_PAVAR)) {
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #13", "%u", cmd->arg1)) break;
        ;       if  ((cmlen > 1) && cmd->arg1) {
        ;           MSG(2, "arg1 in %s[%d] is %d but command length %"BLIN_U" > 1", cmd->arg1, cmlen);
        ;           BREAKORNOT(BSDFIG_EXLEI);
        ;   }   }
        ;   if  (  cmd->arg1
                &&!( ( BSDFIG_ARGTY(BSDFIG_A1NUM)
                     | BSDFIG_ARGTY(BSDFIG_A1BIT)
                     | BSDFIG_ARGTY(BSDFIG_A163K)
                     | BSDFIG_ARGTY(BSDFIG_A1255)
                     | BSDFIG_ARGTY(BSDFIG_A1256)
                     | BSDFIG_ARGTY(BSDFIG_A1127)
                     | BSDFIG_ARGTY(BSDFIG_A1FIB)
                     | BSDFIG_ARGTY(BSDFIG_A1_3)
                     | BSDFIG_ARGTY(BSDFIG_A1_BB)
                     | BSDFIG_ARGTY(BSDFIG_PAVAR)
                     )
                   & SYNCO(opcode).P
                )  ) {
        ;       /* arg1   0,   0 */
        ;       MSG(2, "arg1 in %s[%d] is %d <> %d", cmd->arg1, 0);
        ;       TOBREAKORNOT(BSDFIG_EXLEI);
        ;   }
        ;   if  (cmd->arg1 && elabor(elprinto, 0, "bsdfig_ckout_ipfwrule #14", "[%04X]", cmd->arg1))
                break;
        ;   if  (STYLAR2(BSDFIG_PAIRS, BSDFIG_PAVAR)) {
                size_t i;
        ;       for (i = 1; i < cmlen; i++) {
        ;           if  (elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #15"
                               , "%u..%u", s[i] & 0xFFFF, (s[i] >> 16) & 0xFFFF
                        )      ) break;
        ;       }
        ;       if  (!DOIT(ex)) break;
        ;   }
        ;   if  (  (STYLEIS(BSDFIG_WFULL))
                && cmlen > 1
                && elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #16", "%u", s[1])
                ) break;
        ;   if  (  (STYLEIS(BSDFIG_FLOAT))
                && elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #17"
                         , "%f", (1.0 * s[1]) / 0x7FFFFFFF
                )        ) break;
        ;   if  (  (STYLEIS(BSDFIG_WIP4N))
                && elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #18", "%u.%u.%u.%u", I4N(s[1]))
                ) break;
        ;   if  (  STYLEIS(BSDFIG_WIP4H)
                && elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #19", "%u.%u.%u.%u", I4H(s[1]))
                ) break;
        ;   if  (STYLEIS(BSDFIG_IFA4)) {
        ;       if  (  !((ipfw_insn_if*)cmd)->name[0]) {
                    u_int32_t u;
                    size_t i;
        ;           if  (elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #20"
                               , "%u.%u.%u.%u", I4N(s[1])
                        )      ) break;
        ;           for (i = 0, u = 0; i < IFNAMSIZ; i++) u |= ((ipfw_insn_if*)cmd)->name[i];
        ;           if  (u) TOBREAKORNOT(BSDFIG_EXLEW);
        ;       } else {
        ;           if  (elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #21", "%d", s[1])) break;
#if __FreeBSD_version >= 500000
        ;           if  (s[1] > 1) {
        ;               MSG(2, "arg in %s[%d] is %d > %d", s[1], 1);
        ;               BREAKORNOT(BSDFIG_EXLEI);
        ;           }
#endif
        ;   }   }
        ;   if  (  (STYLEIS(BSDFIG_WIP6N))
                && elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #22"
                         , "%X:%X:%X:%X:%X:%X:%X:%X"
                         , ntohs(s[1] & 0xFFFF), ntohs((s[1] >> 16) & 0xFFFF)
                         , ntohs(s[2] & 0xFFFF), ntohs((s[2] >> 16) & 0xFFFF)
                         , ntohs(s[3] & 0xFFFF), ntohs((s[3] >> 16) & 0xFFFF)
                         , ntohs(s[4] & 0xFFFF), ntohs((s[4] >> 16) & 0xFFFF)
                )        ) break;
        ;   if  (  (STYLEIS(BSDFIG_ETHER))
                && elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #23"
                         , "%02X:%02X:%02X:%02X:%02X:%02X %02X:%02X:%02X:%02X:%02X:%02X "
                           "%02X:%02X:%02X:%02X:%02X:%02X %02X:%02X:%02X:%02X:%02X:%02X"
                         , s[1] & 0xFF, (s[1] >> 8) & 0xFF, (s[1] >> 16) & 0xFF, (s[1] >> 24) & 0xFF
                         , s[2] & 0xFF, (s[2] >> 8) & 0xFF, (s[2] >> 16) & 0xFF, (s[2] >> 24) & 0xFF
                         , s[3] & 0xFF, (s[3] >> 8) & 0xFF, (s[3] >> 16) & 0xFF, (s[3] >> 24) & 0xFF
                         , s[4] & 0xFF, (s[4] >> 8) & 0xFF, (s[4] >> 16) & 0xFF, (s[4] >> 24) & 0xFF
                         , s[5] & 0xFF, (s[5] >> 8) & 0xFF, (s[5] >> 16) & 0xFF, (s[5] >> 24) & 0xFF
                         , s[6] & 0xFF, (s[6] >> 8) & 0xFF, (s[6] >> 16) & 0xFF, (s[6] >> 24) & 0xFF
                )        ) break;
        ;   if  (STYLEIS(BSDFIG_SOCK)) {
                struct sockaddr_in *a;

        ;       a = &((ipfw_insn_sa*)cmd)->sa;
        ;       if  (elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #24"
                           , "%u %u %u.%u.%u.%u"
                           , a->sin_family, a->sin_port, I4N(a->sin_addr.s_addr)
                    )      ) break;
        ;   }
        ;   if  (STYLEIS(BSDFIG_LIMIT)) {
        ;       if  (((ipfw_insn_limit*)cmd)->_pad) {
        ;           MSG(2, "_pad in %s[%d] is %d <> %d", ((ipfw_insn_limit*)cmd)->_pad, 0);
        ;           TOBREAKORNOT(BSDFIG_EXLEI);
        ;       }
        ;       if  (elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #25"
                           , "%02X", ((ipfw_insn_limit*)cmd)->limit_mask
                    )      ) break;
        ;       if  (elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #26"
                           , "%u", ((ipfw_insn_limit*)cmd)->conn_limit
                    )      ) break;
        ;   }
        ;   if  (STYLEIS(BSDFIG_WORDS)) {
                size_t i;
        ;       for (i = 1; i < cmlen; i++) {
        ;           if  (elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #27", "%u", s[i])) break;
        ;       }
        ;       if  (!DOIT(ex)) break;
        ;   }
        ;   if  (STYLEIS(BSDFIG_TEXT0)) {
       /* XXXX     escape */
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #28", "%s", (char*)&s[1])) break;
        ;   }
        ;   if  (STYLEIS(BSDFIG_TEXT1)) {
       /* XXXX     escape */
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #29", "%s", (char*)&s[2])) break;
        ;   }
        ;   if  (STYLAR2(BSDFIG_BITS0, BSDFIG_BITS1)) {
                size_t i = 1; /* for BSDFIG_BITS0 */

        ;       if  (STYLEIS(BSDFIG_BITS1)) i = 2;
        ;       if  (elabor(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #30", NULL)) break;
        ;       for (; i < cmlen; i++) {
        ;           if  (elprint( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #31"
                                , "%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u"
                                ,  s[i]        & 1, (s[i] >>  1) & 1, (s[i] >>  2) & 1, (s[i] >>  3) & 1
                                , (s[i] >>  4) & 1, (s[i] >>  5) & 1, (s[i] >>  6) & 1, (s[i] >>  7) & 1
                                , (s[i] >>  8) & 1, (s[i] >>  9) & 1, (s[i] >> 10) & 1, (s[i] >> 11) & 1
                                , (s[i] >> 12) & 1, (s[i] >> 13) & 1, (s[i] >> 14) & 1, (s[i] >> 15) & 1
                                , (s[i] >> 16) & 1, (s[i] >> 17) & 1, (s[i] >> 18) & 1, (s[i] >> 19) & 1
                                , (s[i] >> 20) & 1, (s[i] >> 21) & 1, (s[i] >> 22) & 1, (s[i] >> 23) & 1
                                , (s[i] >> 24) & 1, (s[i] >> 25) & 1, (s[i] >> 26) & 1, (s[i] >> 27) & 1
                                , (s[i] >> 28) & 1, (s[i] >> 29) & 1, (s[i] >> 30) & 1, (s[i] >> 31) & 1
                        )       ) break;
        ;       }
        ;       if  (!DOIT(ex)) break;
        ;       i = 0;
        ;   }
        ;   if  (STYLEIS(BSDFIG_IP4MS)) {
                size_t i;
        ;       for (i = 1; (i + 2) <= cmlen; i += 2) {
        ;           if  (elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #32"
                               , "%u.%u.%u.%u&%u.%u.%u.%u", I4N(s[i]), I4N(s[i + 1])
                        )      ) break;
                }
        ;       if  (!DOIT(ex)) break;
        ;   }
        ;   if  (STYLEIS(BSDFIG_IP6MS)) {
                size_t i;
        ;       for (i = 1; (i + 8) <= cmlen; i += 8) {
        ;           if  (elabor( elprinto, QPRINT, "bsdfig_ckout_ipfwrule #33"
                               , "%X:%X:%X:%X:%X:%X:%X:%X&%X:%X:%X:%X:%X:%X:%X:%X"
                               , ntohs(s[i    ] & 0xFFFF), ntohs((s[i    ] >> 16) & 0xFFFF)
                               , ntohs(s[i + 1] & 0xFFFF), ntohs((s[i + 1] >> 16) & 0xFFFF)
                               , ntohs(s[i + 2] & 0xFFFF), ntohs((s[i + 2] >> 16) & 0xFFFF)
                               , ntohs(s[i + 3] & 0xFFFF), ntohs((s[i + 3] >> 16) & 0xFFFF)
                               , ntohs(s[i + 4] & 0xFFFF), ntohs((s[i + 4] >> 16) & 0xFFFF)
                               , ntohs(s[i + 5] & 0xFFFF), ntohs((s[i + 5] >> 16) & 0xFFFF)
                               , ntohs(s[i + 6] & 0xFFFF), ntohs((s[i + 6] >> 16) & 0xFFFF)
                               , ntohs(s[i + 7] & 0xFFFF), ntohs((s[i + 7] >> 16) & 0xFFFF)
                        )      ) break;
                }
        ;       if  (!DOIT(ex)) break;
        ;   }
        ;   if  (cmlen > 1) {
                size_t i;

        ;       if  (elabor(elprinto, 0, "bsdfig_ckout_ipfwrule #34", "[")) break;
        ;       ELCONT;
        ;       for (i = 1; i < cmlen; i++) {
        ;           if  (elabor(elprinto, 0, "bsdfig_ckout_ipfwrule #35", "%08X", s[i])) break;
        ;       }
        ;       if  (!DOIT(ex)) break;
        ;       if  (elprint(elprinto, 0, "bsdfig_ckout_ipfwrule #36", "]")) break;
        ;   }
        ;   ELSTOP(QPRINT, "bsdfig_ckout_ipfwrule #37");
        ;   if  ((F_OR & cmd->len) && elprint(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #37", "|"))
                break;
        }
        if  (!(ACTION & flocal)) {
            ifBLIN_QV2(cnf->flags) warnx("bsdfig: not found act_ofs in rule @ %"BLIN_U, rupnt);
            BREAKORNOT(BSDFIG_EXLEE);
        }
        if  (elprint( elprinto, 0, "bsdfig_ckout_ipfwrule #38"
                    , " [%08X %08X %u %u %05u %02u %02X %llu %llu %u]"
#if __FreeBSD_version >= 800504
                    , rule->x_next
#else
                    , rule->next
#endif
                    , rule->next_rule
                    , rule->act_ofs, rule->cmd_len, rule->rulenum, rule->set, rule->_pad
                    , rule->pcnt, rule->bcnt, rule->timestamp
            )       ) break;
        if  (elprint(elprinto, QPRINT, "bsdfig_ckout_ipfwrule #39", ")\n")) break;
    } while (0);
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_ckout_ipfwrule %u\n", ex);
    return(ex);
}

int
bsdfig_ckout_ipfwrules(bsdfig_conf *cnf, bsdfig_ipfwrules *rules) {
    size_t rulen, rupnt;
    u_int32_t flocal = 0;
    struct ip_fw *rule;
    unsigned ex = EX_OK;
    ELPRINTO(cnf, &ex, &flocal);

    if  (!cnf) errx(EX_USAGE, "bsdfig_ckout_ipfwrules: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ckout_ipfwrules %08X\n", cnf->flags);
    if  (!((BSDFIG_PRUFW & cnf->flags) && (cnf->out))) flocal |= DSABLE;
    ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
    if  (!rules) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_ckout_ipfwrules: no rules");
        ex = EX_DATAERR;
        errno = EINVAL;
    }
    for ( rule = rules->rules, rupnt = 0
        ; DOIT(ex) && (rupnt < rules->size)
        ; rule = (struct ip_fw *)&(rule->cmd[rulen]), rupnt += BSDFIG_RULENHEAD + rulen
        ) {
   /*   Loop on every rule */

        if  (SRUEND & flocal) break;
        flocal &= ~(OPRINT | RPRINT | ACTION);
        if  (rupnt + BSDFIG_RULENHEAD > rules->size) {
            ifBLIN_QV1(cnf->flags) warnx("bsdfig_ckout_ipfwrules: short rule @ %"BLIN_U, rupnt);
            TOBREAK(EX_DATAERR);
        }
        rulen = rule->cmd_len;
        if  (rupnt + BSDFIG_RULENHEAD + rulen > rules->size) {
            ifBLIN_QV1(cnf->flags)
                warnx( "bsdfig_ckout_ipfwrules: no space for %"BLIN_U" commands @ %"BLIN_U
                     , rulen, rupnt
                     );
            TOBREAK(EX_DATAERR);
        }
        if  (!DOIT((ex = bsdfig_ckout_ipfwrule(cnf, rule)))) break;
        if  ((rule->rulenum == IPFW_DEFAULT_RULE) && (rule->set == RESVD_SET)) flocal |= SRUEND;
    }
    /*************************************************************************************************
     *                                                                                               *
     *                                                        *
     *                                                                                               *
     *************************************************************************************************/
    if  (BSDFIG_PRDFW & cnf->flags) {
        ipfw_dyn_rule *ruld;
        u_int32_t opcode;

        flocal = 0;
        if  (!((BSDFIG_PRDFW & cnf->flags) && (cnf->out))) flocal |= DSABLE;
        ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
        for ( ruld = (ipfw_dyn_rule*)rule
            ; DOIT(ex) && (rupnt < rules->size)
            ; ruld += 1, rupnt += sizeof(ipfw_dyn_rule) / sizeof(ipfw_insn)
            ) {
        ;   flocal &= ~OPRINT;
        ;   if  (rupnt + (sizeof(ipfw_dyn_rule) / sizeof(ipfw_insn)) > rules->size) {
        ;       ifBLIN_QV1(cnf->flags)
                    warnx("bsdfig_ckout_ipfwrules: no space for rule @ %"BLIN_U, rupnt);
        ;       TOBREAK(EX_DATAERR);
        ;   }
        ;   if  ((opcode = ruld->dyn_type) >= O_LAST_OPCODE) {
        ;       opcode = O_LAST_OPCODE;
        ;       ifBLIN_QV2(cnf->flags)
                    warnx( "bsdfig_ckout_ipfwrules #2: illegal opcode %d >= LAST_OPCODE %d\n"
                         , ruld->dyn_type
                         , O_LAST_OPCODE
                         );
        ;       TOBREAKORNOT(BSDFIG_EXLES);
        ;   }
        ;   if  (elprint( elprinto, QPRINT, "bsdfig_ckout_ipfwrules #40"
                        , "(%05u (", *(u_int16_t*)&ruld->rule
                )       ) break;
        ;   if  (elflowid(elprinto, "bsdfig_ckout_ipfwrules #41", &(ruld->id))) break;
        ;   if  (elprint( elprinto, QPRINT, "bsdfig_ckout_ipfwrules #42"
                        , ") %ds %08X %08X %08X %s"
                        , ruld->expire
                        , ruld->state
                        , ruld->ack_fwd
                        , ruld->ack_rev
                        , SYNCO(opcode).opn
                )       ) break;
        ;   if  (elprint( elprinto, 0, "bsdfig_ckout_ipfwrules #43"
                        , " [%08X %05u %08X %llu %llu ("
                        , ruld->next
                        , *(u_int16_t*)&ruld->rule
                        , ruld->parent
                        , ruld->pcnt
                        , ruld->bcnt
                )       ) break;
        ;   if  (elflowidd(elprinto, "bsdfig_ckout_ipfwrules #44", &(ruld->id))) break;
        ;   if  (elprint( elprinto, 0, "bsdfig_ckout_ipfwrules #45"
                        , " ) %u %u %08X %08X %08X %04X %u]", &flocal
/* XXXX   &flocal.      */
                        , ruld->expire
                        , ruld->bucket
                        , ruld->state
                        , ruld->ack_fwd
                        , ruld->ack_rev
                        , ruld->dyn_type
                        , ruld->count
                )       ) break;
        ;   if  (elprint( elprinto, QPRINT, "bsdfig_ckout_ipfwrules #46", ")\n")) break;
    }   }
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_ckout_ipfwrules %u\n", ex);
    return(ex);
}

int
bsdfig_out_ipfwset(bsdfig_conf *cnf, bsdfig_ipfwrules *rules) {
    u_int32_t flocal = 0, s;
    unsigned ex = EX_OK;

    if  (!cnf) errx(EX_USAGE, "bsdfig_out_ipfwset: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_out_ipfwset %08X\n", cnf->flags);
    if  (!((BSDFIG_PRSET & cnf->flags) && (cnf->out))) flocal |= DSABLE;
    ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
    if  (!rules) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_out_ipfwset: no rules");
        ex = EX_DATAERR;
        errno = EINVAL;
    }
    if  (!ex && (  (BSDFIG_RULENHEAD > rules->size)
                || (BSDFIG_RULENHEAD + rules->rules[0].cmd_len > rules->size)
        )       ) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_out_ipfwset: short rules");
        ex = EX_DATAERR;
        errno = EINVAL;
    }
    s = BLIN_I(rules->rules[0].next_rule);
    if  (  !ex
        && !(DSABLE  & flocal)
        && (fprintf( cnf->out
                   , "DISABLE(%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u%u)\n"
                   ,  s        & 1, (s >>  1) & 1, (s >>  2) & 1, (s >>  3) & 1
                   , (s >>  4) & 1, (s >>  5) & 1, (s >>  6) & 1, (s >>  7) & 1
                   , (s >>  8) & 1, (s >>  9) & 1, (s >> 10) & 1, (s >> 11) & 1
                   , (s >> 12) & 1, (s >> 13) & 1, (s >> 14) & 1, (s >> 15) & 1
                   , (s >> 16) & 1, (s >> 17) & 1, (s >> 18) & 1, (s >> 19) & 1
                   , (s >> 20) & 1, (s >> 21) & 1, (s >> 22) & 1, (s >> 23) & 1
                   , (s >> 24) & 1, (s >> 25) & 1, (s >> 26) & 1, (s >> 27) & 1
                   , (s >> 28) & 1, (s >> 29) & 1, (s >> 30) & 1, (s >> 31) & 1
                   ) < 0
        )  ) {
        ifBLIN_QV1(cnf->flags) warn("bsdfig_out_ipfwset: fprintf");
        ex = EX_IOERR;
    }
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_out_ipfwset %u\n", ex);
    return(ex);
}

#if __FreeBSD_version >= 700000
int
bsdfig_ckout_nats(bsdfig_conf *cnf, bsdfig_ipfwnats *nats) {
    int natc;
    u_int32_t flocal = 0;
    unsigned ex = EX_OK;
    struct cfg_nat *nat;
    size_t npnt;
    ELPRINTO(cnf, &ex, &flocal);

    if  (!cnf) errx(EX_USAGE, "bsdfig_ckout_ipfwnats: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ckout_ipfwnats %08X\n", cnf->flags);
    ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
    if  (!nats) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_ckout_nats: no nats");
        ex = EX_DATAERR;
        errno = EINVAL;
    }
    for (npnt = 0, natc = 0; DOIT(ex) && natc < nats->nat_cnt ; natc++) {
        struct cfg_redir *rat;
        int redirc, i;

        nat = (struct cfg_nat *)((char*)(nats->nats) + npnt);
        npnt += sizeof(struct cfg_nat);
        if  (npnt > nats->size) {
            ifBLIN_QV1(cnf->flags)
                warnx( "bsdfig_ckout_nats: natN %u < natC %u but P %"BLIN_U" > SZ %"BLIN_U
                     , natc, nats->nat_cnt
                     , npnt, nats->size
                     );
            TOBREAKORNOT(BSDFIG_EXLES);
        }
        if  (elprint( elprinto, QPRINT, "ckout_nats #0"
                    , "NAT(%d %u.%u.%u.%u %d"
                    , nat->id, I4N(nat->ip.s_addr), nat->mode
            )       ) break;
        if  (elprint( elprinto, QPRINT, "ckout_nats #m"
                    , "[%s%s%s%s%s%s%s%s%s ]"
                    , (nat->mode & PKT_ALIAS_LOG                 ) ? " LOG"                  : ""
                    , (nat->mode & PKT_ALIAS_DENY_INCOMING       ) ? " DENY_INCOMING"        : ""
                    , (nat->mode & PKT_ALIAS_SAME_PORTS          ) ? " SAME_PORTS"           : ""
                    , (nat->mode & PKT_ALIAS_USE_SOCKETS         ) ? " USE_SOCKETS"          : ""
                    , (nat->mode & PKT_ALIAS_UNREGISTERED_ONLY   ) ? " UNREGISTERED_ONLY"    : ""
                    , (nat->mode & PKT_ALIAS_RESET_ON_ADDR_CHANGE) ? " RESET_ON_ADDR_CHANGE" : ""
                    , (nat->mode & PKT_ALIAS_PROXY_ONLY          ) ? " PROXY_ONLY"           : ""
                    , (nat->mode & PKT_ALIAS_REVERSE             ) ? " REVERSE"              : ""
                    , (nat->mode & PKT_ALIAS_PUNCH_FW            ) ? " PUNCH_FW"             : ""
            )       ) break;
        if  (elprint(elprinto, QPRINT, "ckout_nats #n", " ")) break;
        for (i = 0; i < IF_NAMESIZE; i++) {
            if  (!nat->if_name[i]) break;
            if  (elprint( elprinto, QPRINT, "ckout_nats #1", "%c", nat->if_name[i])) break;
        }
        if  (elprint( elprinto, 0, "ckout_nats #2"
                    , " [%"BLIN_X" %"BLIN_X" %08X '"
                    , BLIN_I(nat->_next.le_next), BLIN_I(nat->_next.le_prev), nat->ip.s_addr
            )       ) break;
        for (i = 0; i < IF_NAMESIZE; i++) {
            if  (nat->if_name[i] < ' ') {
                if  (elprint( elprinto, 0, "ckout_nats #3", "\\%03o", nat->if_name[i])) break;
            } else {
                if  (elprint( elprinto, 0, "ckout_nats #9", "%c", nat->if_name[i])) break;
        }   }
        if  (elprint( elprinto, 0, "ckout_nats #4"
                    , "' %d %"BLIN_X" %d %"BLIN_X"]"
                    , nat->mode, nat->lib, nat->redir_cnt, nat->redir_chain.lh_first
            )       ) break;
        for (redirc = 0; redirc < nat->redir_cnt; redirc++, rat++) {
            struct cfg_spool *sat;
            u_int16_t spoolc;

            rat = (struct cfg_redir *)((char*)(nats->nats) + npnt);
            npnt += sizeof(struct cfg_redir);
            if  (npnt > nats->size) {
                ifBLIN_QV1(cnf->flags)
                    warnx( "bsdfig_ckout_nats: redirN %u < redirC %u but P %"BLIN_U" > SZ %"BLIN_U
                         , redirc, nat->redir_cnt
                         , npnt, nats->size
                         );
                TOBREAKORNOT(BSDFIG_EXLES);
            }
            if  (elprint( elprinto, QPRINT, "ckout_nats #5"
                        , "\n  REDIRECT(%04X %u.%u.%u.%u %u.%u.%u.%u %u.%u.%u.%u %u %u %u %u %u %d"
                        , rat->mode
                        , I4N(rat->laddr.s_addr), I4N(rat->paddr.s_addr), I4N(rat->raddr.s_addr)
                        , rat->lport, rat->pport, rat->rport
                        , rat->pport_cnt, rat->rport_cnt, rat->proto
                )       ) break;
            if  (elprint( elprinto, 0, "ckout_nats #6"
                        , " [%"BLIN_X" %"BLIN_X" %04X %08X %08X %08X %"BLIN_X" %u %"BLIN_X"]"
                        , BLIN_I(rat->_next.le_next), BLIN_I(rat->_next.le_prev), rat->mode
                        , rat->laddr.s_addr, rat->paddr.s_addr, rat->raddr.s_addr
                        , rat->alink, rat->spool_cnt, rat->spool_chain.lh_first
                )       ) break;
            for (spoolc = 0; spoolc < rat->spool_cnt; spoolc++, sat++) {
                sat = (struct cfg_spool *)((char*)(nats->nats) + npnt);
                npnt += sizeof(struct cfg_spool);
                if  (npnt > nats->size) {
                    ifBLIN_QV1(cnf->flags)
                        warnx( "bsdfig_ckout_nats: spoolN %u < spoolC %u but P %"BLIN_U" > SZ %"BLIN_U
                             , spoolc, rat->spool_cnt
                             , npnt, nats->size
                             );
                    TOBREAKORNOT(BSDFIG_EXLES);
                }
                if  (elprint( elprinto, QPRINT, "ckout_nats #7"
                            , "\n    SPOOL(%u.%u.%u.%u %u"
                            , I4N(sat->addr.s_addr), sat->port
                    )       ) break;
                if  (elprint( elprinto, 0, "ckout_nats #8"
                            , " [%"BLIN_X" %"BLIN_X" %08X]"
                            , BLIN_I(sat->_next.le_next), BLIN_I(sat->_next.le_prev), sat->addr.s_addr
                    )       ) break;
                if  (elprint(elprinto, QPRINT, "ckout_nats #-7", ")")) break;
            }
            if  (rat->spool_cnt) {
                if  (elprint(elprinto, QPRINT, "ckout_nats #-6", "\n  )")) break;
            } else {
                if  (elprint(elprinto, QPRINT, "ckout_nats #-5", ")")) break;
            }
            rat = (struct cfg_redir *)((char*)(nats->nats) + npnt);
        }
        if  (nat->redir_cnt) {
            if  (elprint(elprinto, QPRINT, "ckout_nats #-1", "\n)\n")) break;
        } else {
            if  (elprint(elprinto, QPRINT, "ckout_nats #-0", ")\n")) break;
    }   }
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_out_ipfwnats %u\n", ex);
    return(ex);
}

int
bsdfig_ckout_natl(bsdfig_conf *cnf, bsdfig_ipfwnatl *natl) {
    size_t nnum, npnt;
    u_int32_t flocal = 0;
    unsigned ex = EX_OK;
    void *nat;
    ELPRINTO(cnf, &ex, &flocal);

    if  (!cnf) errx(EX_USAGE, "bsdfig_ckout_ipfwnatl: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ckout_ipfwnatl %08X\n", cnf->flags);
    ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
    if  (!natl) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_ckout_natl: no natl");
        ex = EX_DATAERR;
        errno = EINVAL;
    }
    for (npnt = 0; DOIT(ex); ) {
        nat = (void*)((char*)(natl->natl) + npnt);
        npnt += sizeof(int);
        if  (npnt > natl->size) {
            ifBLIN_QV1(cnf->flags)
                warnx( "bsdfig_ckout_natl #1: P %"BLIN_U" > SZ %"BLIN_U
                     , npnt, natl->size
                     );
            TOBREAKORNOT(BSDFIG_EXLES);
        }
        bcopy(nat, &nnum, sizeof(int));
        if  (npnt == natl->size) {
            if  (elprint(elprinto, 0, "bsdfig_ckout_natl #1", "(no nat %u)\n", nnum)) break;
            break;
        }
        nat = (void*)((char*)(natl->natl) + npnt);
        npnt += LIBALIAS_BUF_SIZE;
        if  (npnt > natl->size) {
            ifBLIN_QV1(cnf->flags)
                warnx( "bsdfig_ckout_natl #2: P %"BLIN_U" > SZ %"BLIN_U
                     , npnt, natl->size
                     );
            TOBREAKORNOT(BSDFIG_EXLES);
        }
        if  (elprint( elprinto, QPRINT, "bsdfig_ckout_natl #1", "(nat %u: %s)\n"
                    , nnum, (char*)nat
            )       ) break;
    }
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_out_ipfwnatl %u\n", ex);
    return(ex);
}
#endif

static int
ckout_tablentry(bsdfig_conf *cnf, u_int32_t *flocal, ipfw_table_entry *e) {
    unsigned ex = EX_OK;
    ELPRINTO(cnf, &ex, flocal);

    elprint( elprinto, QPRINT, "ckout_tables #1"
           , "TABLE(%u %u %u.%u.%u.%u/%u"
           , e->tbl, e->value, I4N(e->addr), e->masklen
           );
    elprint( elprinto, 0, "ckout_tables #2"
           , " [%08X %u %u %u]"
           , e->addr, e->value, e->tbl, e->masklen
           );
    elprint(elprinto, QPRINT, "ckout_tables #3", ")\n");
    return(ex);
}

int
bsdfig_ckout_ipfwtables(bsdfig_conf *cnf, bsdfig_ipfwtables *tables) {
    u_int32_t flocal = 0;
    unsigned ex = EX_OK;
    size_t i, j;
    ELPRINTO(cnf, &ex, &flocal);

    if  (!cnf) errx(EX_USAGE, "bsdfig_ckout_ipfwtables: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ckout_ipfwtables %08X\n", cnf->flags);
    if  (!((BSDFIG_FTABL & cnf->flags) && (cnf->out))) flocal |= DSABLE;
    ifBLIN_QV3(cnf->flags) flocal |= FPRINT;
    for (i = 0; !ex && (i < tables->tabn); i++) if (tables->tables[i]) {
        elprint( elprinto, 0, "bsdfig_ckout_ipfwtables", "[%u %u %u]\n"
               , tables->tables[i]->size
               , tables->tables[i]->cnt
               , tables->tables[i]->tbl
               );
        for (j = 0; !ex && (j < tables->tables[i]->cnt); j++) {
            if  (j >= tables->tables[i]->size) {
                ifBLIN_QV1(cnf->flags) warnx("bsdfig_ckout_ipfwtables: table %"BLIN_U" shorten", j);
                errno = EMSGSIZE;
                ex = EX_SOFTWARE;
                break;
            }
            ex = ckout_tablentry(cnf, &flocal, &(tables->tables[i]->ent[j]));
    }   }
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_ckout_ipfwtables %u\n", ex);
    return(ex);
}

int
bsdfig_pipelist(bsdfig_conf *cnf, u_int64_t *larr) {
    unsigned ex = EX_OK;

    if  (!cnf) errx(EX_USAGE, "bsdfig_pipelist: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_pipelist %08X\n", cnf->flags);
    ex = (larr)
       ? elist(cnf, larr, BSDFIG_IPFW_OTHERS_MAX)
       : elist(cnf, cnf->conf->my_pipes, BSDFIG_IPFW_OTHERS_MAX);
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_pipelist %u\n", ex);
    return(ex);
}

int
bsdfig_queuelist(bsdfig_conf *cnf, u_int64_t *larr) {
    unsigned ex = EX_OK;

    if  (!cnf) errx(EX_USAGE, "bsdfig_queuelist: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_queuelist %08X\n", cnf->flags);
    ex = (larr)
       ? elist(cnf, larr, BSDFIG_IPFW_OTHERS_MAX)
       : elist(cnf, cnf->conf->my_queues, BSDFIG_IPFW_OTHERS_MAX);
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_queuelist %u\n", ex);
    return(ex);
}

int
bsdfig_disablelist(bsdfig_conf *cnf, u_int64_t *larr) {
    unsigned ex = EX_OK;

    if  (!cnf) errx(EX_USAGE, "bsdfig_disablelist: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_disablelist %08X\n", cnf->flags);
    ex = (larr)
       ? elist(cnf, larr, BSDFIG_IPFW_DISABLES_MAX)
       : elist(cnf, cnf->conf->my_disables, BSDFIG_IPFW_DISABLES_MAX);
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_disablelist %u\n", ex);
    return(ex);
}

int
bsdfig_ipfwrulelist(bsdfig_conf *cnf, u_int64_t *larr) {
    unsigned ex = EX_OK;

    if  (!cnf) errx(EX_USAGE, "bsdfig_ipfwrulelist: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ipfwrulelist %08X\n", cnf->flags);
    ex = (larr)
       ? elist(cnf, larr, BSDFIG_IPFW_OTHERS_MAX)
       : elist(cnf, cnf->conf->my_ipfwrules, BSDFIG_IPFW_OTHERS_MAX);
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_ipfwrulelist %u\n", ex);
    return(ex);
}

#if __FreeBSD_version >= 700000
int
bsdfig_ipfwnatslist(bsdfig_conf *cnf, u_int64_t *larr) {
    unsigned ex = EX_OK;

    if  (!cnf) errx(EX_USAGE, "bsdfig_ipfwnatslist: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ipfwnatslist %08X\n", cnf->flags);
    ex = (larr)
       ? elist(cnf, larr, BSDFIG_IPFW_OTHERS_MAX)
       : elist(cnf, cnf->conf->my_ipfwnats, BSDFIG_IPFW_OTHERS_MAX);
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_ipfwnatslist %u\n", ex);
    return(ex);
}
#endif

int
bsdfig_ipfwtablelist(bsdfig_conf *cnf, u_int64_t *larr) {
    unsigned ex = EX_OK;

    if  (!cnf) errx(EX_USAGE, "bsdfig_ipfwtablelist: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_ipfwtablelist %08X\n", cnf->flags);
    ex = (larr) ? elist(cnf, larr, cnf->tabn) : elist(cnf, cnf->conf->my_ipfwtables, cnf->conf->tabn);
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_ipfwtable %u\n", ex);
    return(ex);
}
