/*-
 * Copyright (C) @BABOLO  2009 May 08 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2009 May 08 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: ipfw_syntax.c,v 1.15 2011/11/26 12:29:09 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <stdio.h>
#include <mife.h>
#include <multilar.h>
#include "bsdfig.h"

#if __FreeBSD_version < 500000
# include <netinet/ip_fw2.h>
# include <net/route.h>
# include <sys/mbuf.h>
#else
# include <netinet/ip_fw.h>
#endif

#define LM    BSDFIG_LMDLM  /*    */
#define LI    BSDFIG_LMDLI  /*   1 + 8n   */

#define Ap    BSDFIG_ROLAp
#define As    BSDFIG_ROLAs
#define Ac    BSDFIG_ROLAc
#define Ak    BSDFIG_ROLAk
#define Al    BSDFIG_ROLAl
#define Aq    BSDFIG_ROLAq
#define At    BSDFIG_ROLAt
#define Aa    BSDFIG_ROLAa
#define An    BSDFIG_ROLAn
#define Ms    0
#define nX    BSDFIG_ROLpl

#define Hn BSDFIG_ARGTY(BSDFIG_A1NUM)
#define HN BSDFIG_ARGTY(BSDFIG_A1BIT)
#define Hh BSDFIG_ARGTY(BSDFIG_A163K)
#define H8 BSDFIG_ARGTY(BSDFIG_A1255)
#define H9 BSDFIG_ARGTY(BSDFIG_A1256)
#define Ht BSDFIG_ARGTY(BSDFIG_A1127)
#define Hr BSDFIG_ARGTY(BSDFIG_A1FIB)
#define H2 BSDFIG_ARGTY(BSDFIG_A1_3)
#define Hb BSDFIG_ARGTY(BSDFIG_A1_BB)
#define Vv BSDFIG_ARGTY(BSDFIG_PAVAR)
#define Wh BSDFIG_ARGTY(BSDFIG_WFULL)
#define Wf BSDFIG_ARGTY(BSDFIG_FLOAT)
#define Wi BSDFIG_ARGTY(BSDFIG_WIP4N)
#define Wj BSDFIG_ARGTY(BSDFIG_WIP4H)
#define Wa BSDFIG_ARGTY(BSDFIG_IFA4)
#define W6 BSDFIG_ARGTY(BSDFIG_WIP6N)
#define We BSDFIG_ARGTY(BSDFIG_ETHER)
#define Ws BSDFIG_ARGTY(BSDFIG_SOCK)
#define Wl BSDFIG_ARGTY(BSDFIG_LIMIT)
#define Vp BSDFIG_ARGTY(BSDFIG_PAIRS)
#define Vh BSDFIG_ARGTY(BSDFIG_WORDS)
#define Vt BSDFIG_ARGTY(BSDFIG_TEXT0)
#define VT BSDFIG_ARGTY(BSDFIG_TEXT1)
#define Vb BSDFIG_ARGTY(BSDFIG_BITS0)
#define VB BSDFIG_ARGTY(BSDFIG_BITS1)
#define V4 BSDFIG_ARGTY(BSDFIG_IP4MS)
#define V6 BSDFIG_ARGTY(BSDFIG_IP6MS)
#define Er BSDFIG_ARGTY(BSDFIG_ERROR)

static const bsdfig_syllable Wone[] = {0xFFFF0000, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
static const bsdfig_syllable Wtre[] = {0, 0, 0xFFFFFFFF};
static const bsdfig_syllable What[] = {0xFFFC0000};

#if __FreeBSD_version >= 500000
const bsdfig_syntax __bsdfig_s =
#else
const bsdfig_syntax bsdfig_s =
#endif
{ 0001020304050607LLU
, "bsdfig\000S"
, O_LAST_OPCODE
, { {{{ 1,31, 0, Ms|Ac,         Vt}}, Wone, "NOP"          , O_NOP          } /* 00 */
  , {{{ 2, 2, 0, Ms|Ac,         Wi}}, Wone, "IP_SRC"       , O_IP_SRC       }
  , {{{ 3,31,LM, Ms|Ac,         V4}}, Wone, "IP_SRC_MASK"  , O_IP_SRC_MASK  }
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "IP_SRC_ME"    , O_IP_SRC_ME    }
  , {{{ 3,10, 0, Ms|Ac,   HN|Wj|VB}}, NULL, "IP_SRC_SET"   , O_IP_SRC_SET   } /* 04 */
  , {{{ 2, 2, 0, Ms|Ac,         Wi}}, Wone, "IP_DST"       , O_IP_DST       }
  , {{{ 3,31,LM, Ms|Ac,         V4}}, Wone, "IP_DST_MASK"  , O_IP_DST_MASK  }
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "IP_DST_ME"    , O_IP_DST_ME    }
  , {{{ 3,10, 0, Ms|Ac,   HN|Wj|VB}}, NULL, "IP_DST_SET"   , O_IP_DST_SET   } /* 08 */
  , {{{ 2,31, 0, Ms|Ac,         Vp}}, Wone, "IP_SRCPORT"   , O_IP_SRCPORT   }
  , {{{ 2,31, 0, Ms|Ac,         Vp}}, Wone, "IP_DSTPORT"   , O_IP_DSTPORT   }
  , {{{ 1, 1, 0, Ms|Ac,         H8}}, Wone, "PROTO"        , O_PROTO        }
  , {{{ 7, 7, 0, Ms|Ac,         We}}, Wone, "MACADDR2"     , O_MACADDR2     } /* 0C */
  , {{{ 2,31, 0, Ms|Ac,         Vp}}, NULL, "MAC_TYPE"     , O_MAC_TYPE     }
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "LAYER2"       , O_LAYER2       }
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "IN"           , O_IN           }
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "FRAG"         , O_FRAG         } /* 10 */
  , {{{ 6, 6, 0, Ms|Ac,      Wa|VT}}, Wone, "RECV"         , O_RECV         }
  , {{{ 6, 6, 0, Ms|Ac,      Wa|VT}}, Wone, "XMIT"         , O_XMIT         }
  , {{{ 6, 6, 0, Ms|Ac,      Wa|VT}}, Wone, "VIA"          , O_VIA          }
  , {{{ 1, 1, 0, Ms|Ac,   Hb      }}, NULL, "IPOPT"        , O_IPOPT        } /* 14 */
  , {{{ 1,31, 0, Ms|Ac,      Vv   }}, NULL, "IPLEN"        , O_IPLEN        }
  , {{{ 1,31, 0, Ms|Ac,      Vv   }}, NULL, "IPID"         , O_IPID         }
  , {{{ 1, 1, 0, Ms|Ac,   Hb      }}, NULL, "IPTOS"        , O_IPTOS        }
  , {{{ 1, 1, 0, Ms|Ac,   H8      }}, NULL, "IPPRECEDENCE" , O_IPPRECEDENCE } /* 18 */
  , {{{ 1,31, 0, Ms|Ac,      Vv   }}, NULL, "IPTTL"        , O_IPTTL        }
  , {{{ 1, 1, 0, Ms|Ac,   H8      }}, NULL, "IPVER"        , O_IPVER        }
  , {{{ 2, 2, 0, Ms|Ac,         Wh}}, Wtre, "UID"          , O_UID          }
  , {{{ 2, 2, 0, Ms|Ac,         Wh}}, Wtre, "GID"          , O_GID          } /* 1C */
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "ESTAB"        , O_ESTAB        }
  , {{{ 1, 1, 0, Ms|Ac,   Hb      }}, NULL, "TCPFLAGS"     , O_TCPFLAGS     }
  , {{{ 1, 1, 0, Ms|Ac,   Hh      }}, NULL, "TCPWIN"       , O_TCPWIN       }
  , {{{ 2, 2, 0, Ms|Ac,         Wh}}, Wone, "TCPSEQ"       , O_TCPSEQ       } /* 20 */
  , {{{ 2, 2, 0, Ms|Ac,         Wh}}, Wone, "TCPACK"       , O_TCPACK       }
  , {{{ 2, 2, 0, Ms|Ac,         Vb}}, Wtre, "ICMPTYPE"     , O_ICMPTYPE     }
  , {{{ 1, 1, 0, Ms|Ac,   Hb      }}, NULL, "TCPOPTS"      , O_TCPOPTS      }
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "VERREVPATH"   , O_VERREVPATH   } /* 24 */
#if __FreeBSD_version >= 500000
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "VERSRCREACH"  , O_VERSRCREACH  }
#endif
  , {{{ 1, 1, 0, nX|As,          0}}, Wone, "PROBE_STATE"  , O_PROBE_STATE  }
  , {{{ 1, 1, 0, nX|Ak,          0}}, Wone, "KEEP_STATE"   , O_KEEP_STATE   }
  , {{{ 2, 2, 0, nX|Ak,         Wl}}, Wone, "LIMIT"        , O_LIMIT        } /* 28 */
  , {{{ 0,63, 0,     0,Er|Hh   |Vh}}, NULL, "LIMIT_PARENT" , O_LIMIT_PARENT }
  , {{{ 3, 3, 0, nX|Al,         Wh}}, Wtre, "LOG"          , O_LOG          }
  , {{{ 2, 2, 0, nX|Ap,         Wf}}, Wone, "PROB"         , O_PROB         }
  , {{{ 1, 1, 0, nX|Aa,          0}}, Wone, "CHECK_STATE"  , O_CHECK_STATE  } /* 2C */
  , {{{ 1, 1, 0, nX|Aa,          0}}, Wone, "ACCEPT"       , O_ACCEPT       }
  , {{{ 1, 1, 0, nX|Aa,          0}}, Wone, "DENY"         , O_DENY         }
  , {{{ 1, 1, 0, nX|Aa,   Hh      }}, NULL, "REJECT"       , O_REJECT       }
  , {{{ 1, 1, 0, nX|Aa,          0}}, Wone, "COUNT"        , O_COUNT        } /* 30 */
  , {{{ 1, 1, 0, nX|Aa,   Hh      }}, NULL, "SKIPTO"       , O_SKIPTO       }
#if __FreeBSD_version >= 500000
  , {{{ 1, 1, 0, nX|Aa,   Hh      }}, NULL, "PIPE"         , O_PIPE         }
  , {{{ 1, 1, 0, nX|Aa,   Hh      }}, NULL, "QUEUE"        , O_QUEUE        }
#else
  , {{{ 2, 2, 0, nX|Aa,   Hh      }}, NULL, "PIPE"         , O_PIPE         }
  , {{{ 2, 2, 0, nX|Aa,   Hh      }}, NULL, "QUEUE"        , O_QUEUE        }
#endif
  , {{{ 1, 1, 0, nX|Aa,   Hh      }}, NULL, "DIVERT"       , O_DIVERT       } /* 34 */
  , {{{ 1, 1, 0, nX|Aa,   Hh      }}, NULL, "TEE"          , O_TEE          }
  , {{{ 5, 5, 0, nX|Aa,         Ws}}, Wone, "FORWARD_IP"   , O_FORWARD_IP   }
  , {{{ 0,63, 0, nX|Aa,Er|Hh   |Vh}}, NULL, "FORWARD_MAC"  , O_FORWARD_MAC  }
#if __FreeBSD_version >= 700000
  , {{{ 2, 2, 0, nX|Aa,   Hh      }}, NULL, "NAT"          , O_NAT          } /* 38 */
#endif
#if __FreeBSD_version > 800099
  , {{{ 1, 1, 0, nX|Aa,          0}}, Wone, "REASS"        , O_REASS        }
#endif
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "IPSEC"        , O_IPSEC        }
#if __FreeBSD_version >= 800500
  , {{{ 1, 3, 0, Ms|Ac,   Hh   |Vh}}, NULL, "IP_SRC_LOOKUP", O_IP_SRC_LOOKUP}
  , {{{ 1, 3, 0, Ms|Ac,   Hh   |Vh}}, NULL, "IP_DST_LOOKUP", O_IP_DST_LOOKUP} /* 3C */
#else
  , {{{ 1, 2, 0, Ms|Ac,   Hh   |Wh}}, NULL, "IP_SRC_LOOKUP", O_IP_SRC_LOOKUP}
  , {{{ 1, 2, 0, Ms|Ac,   Hh   |Wh}}, NULL, "IP_DST_LOOKUP", O_IP_DST_LOOKUP} /* 3C */
#endif
#if __FreeBSD_version >= 600000
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "ANTISPOOF"    , O_ANTISPOOF    } /* 3C */
  , {{{ 2, 2, 0, Ms|Ac,         Wh}}, Wtre, "JAIL"         , O_JAIL         }
  , {{{ 2, 2, 0, nX|Aq,         Wh}}, Wone, "ALTQ"         , O_ALTQ         }
  , {{{ 1, 1, 0, Ms|Ac,   H2      }}, What, "DIVERTED"     , O_DIVERTED     } /* 40 */
  , {{{ 1,31, 0, Ms|Ac,      Vv   }}, NULL, "TCPDATALEN"   , O_TCPDATALEN   } /* 40 */
  , {{{ 5, 5, 0, Ms|Ac,         W6}}, Wone, "IP6_SRC"      , O_IP6_SRC      }
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "IP6_SRC_ME"   , O_IP6_SRC_ME   }
  , {{{ 9,57,LI, Ms|Ac,         V6}}, Wone, "IP6_SRC_MASK" , O_IP6_SRC_MASK } /* 44 */
  , {{{ 5, 5, 0, Ms|Ac,         W6}}, Wone, "IP6_DST"      , O_IP6_DST      } /* 44 */
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "IP6_DST_ME"   , O_IP6_DST_ME   }
  , {{{ 9,57,LI, Ms|Ac,         V6}}, Wone, "IP6_DST_MASK" , O_IP6_DST_MASK }
  , {{{ 2,63, 0, Ms|Ac,   Hn   |Vh}}, NULL, "FLOW6ID"      , O_FLOW6ID      } /* 48 */
  , {{{ 8, 8, 0, Ms|Ac,         Vb}}, Wone, "ICMP6TYPE"    , O_ICMP6TYPE    } /* 48 */
  , {{{ 1, 1, 0, Ms|Ac,   Hh      }}, NULL, "EXT_HDR"      , O_EXT_HDR      }
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "IP6"          , O_IP6          }
  , {{{ 1, 1, 0, nX|Aa,   Hh      }}, NULL, "NETGRAPH"     , O_NETGRAPH     } /* 4C */
  , {{{ 1, 1, 0, nX|Aa,   Hh      }}, NULL, "NGTEE"        , O_NGTEE        } /* 4C */
  , {{{ 1, 1, 0, Ms|Ac,          0}}, Wone, "IP4"          , O_IP4          }
  , {{{ 1, 1, 0, nX|Aa,   Hh      }}, NULL, "UNREACH6"     , O_UNREACH6     }
  , {{{ 1, 1, 0, nX|At,   Hh      }}, NULL, "TAG"          , O_TAG          } /* 50 */
  , {{{ 1,31, 0, Ms|Ac,      Vv   }}, NULL, "TAGGED"       , O_TAGGED       } /* 50 */
#endif
#if __FreeBSD_version >= 700100
  , {{{ 1, 1, 0, nX|Aa,   Hr      }}, NULL, "SETFIB"       , O_SETFIB       }
  , {{{ 1, 1, 0, Ms|Ac,   Hr      }}, NULL, "FIB"          , O_FIB          }
#endif
  , {{{ 0,63, 0,     0,Er|Hh   |Vh}}, NULL, "!!!"          , O_LAST_OPCODE  }
} };

void
bsdfig_syntaxdump(bsdfig_conf *cnf) {
    int i;
    for (i = 0; i < O_LAST_OPCODE; i++) {
        fprintf( stderr, "%3d%s%02X %016llX %s\n"
               , i, (i == cnf->syntax->p[i].code) ? "==" : "<>", cnf->syntax->p[i].code
               , cnf->syntax->p[i].p
               , cnf->syntax->p[i].opn
               );
    }
    fprintf(stderr, "sizeof(struct dn_pipe)       %4d 0x%04X\n", sizeof(struct dn_pipe), sizeof(struct dn_pipe));
    fprintf(stderr, "sizeof(struct dn_flow_set)   %4d 0x%04X\n", sizeof(struct dn_flow_set), sizeof(struct dn_flow_set));
    fprintf(stderr, "sizeof(struct dn_flow_queue) %4d 0x%04X\n", sizeof(struct dn_flow_queue), sizeof(struct dn_flow_queue));
}

#if __FreeBSD_version >= 500000
__weak_reference(__bsdfig_s, bsdfig_s);
#endif
